MODULE UTILS
!USE NUMERICAL_LIBRARIES
!
IMPLICIT NONE
!
!public :: length
public :: asym
public :: CROSSPROD
public :: CAMAT
public :: SUMROT
public :: SUMROTREL
public :: AMAT2EULER
public :: EV2EP
public :: eaa2ep
public :: ep2eaa
public :: get_two_plane_vectors
public :: vec_length
public :: vec_norm
public :: rotate
public :: rotate_mat
public :: intpol_v
public :: EV2AMAT
!
CONTAINS
!
!********************
!FUNCTION LENGTH(N,V)
!!********************
!! Purpose: Calc. the length of a vector
!! input
!INTEGER          :: N
!DOUBLE PRECISION :: V(N)
!! output
!DOUBLE PRECISION :: LENGTH
!! locals
!!
!LENGTH = DSQRT(DDOT(N,V,1,V,1))
!!
!END FUNCTION LENGTH
!********************

!***********************
SUBROUTINE ASYM(vec,mat)
!***********************
! Purpose: Calc. the ASYM matrix
!
! input
double precision vec(3)
! 
! output
double precision mat(3,3)
mat = RESHAPE((/ 0.d0,   vec(3),  -vec(2), &
                -vec(3),   0.d0,   vec(1), &
				 vec(2), -vec(1),   0.d0 /),(/3,3/))
END SUBROUTINE ASYM
!***********************
!
!******************************
FUNCTION CROSSPROD(v1,v2)
!******************************
! Purpose: Calculates the cross product between 2 vectors
! input
DOUBLE PRECISION    :: v1(3),v2(3)
! output        
DOUBLE PRECISION    :: CROSSPROD(3)
!
CROSSPROD(1) = v1(2) * v2(3) - v1(3) * v2(2)
CROSSPROD(2) = v1(3) * v2(1) - v1(1) * v2(3)
CROSSPROD(3) = v1(1) * v2(2) - v1(2) * v2(1)
!
END FUNCTION CROSSPROD
!
!***************************
SUBROUTINE CAMAT(EULER,AMAT)
!***************************
! Purpose: Calculates the Transformation matrix related to input Euler parameters
! Input
DOUBLE PRECISION    :: EULER(4)        ! Euler parameters
! Output
DOUBLE PRECISION    :: AMAT(3,3)       ! Transformation matrix
!
AMAT = RESHAPE((/&
         1.d0-2.d0*(EULER(3)**2+EULER(4)**2),& 
         2.d0*(EULER(2)*EULER(3)+EULER(1)*EULER(4)),&
		 2.d0*(EULER(2)*EULER(4)-EULER(1)*EULER(3)),&
         2.d0*(EULER(2)*EULER(3)-EULER(1)*EULER(4)),&
		 1.d0-2.d0*(EULER(2)**2+EULER(4)**2), &
		 2.d0*(EULER(3)*EULER(4)+EULER(1)*EULER(2)),&
		 2.d0*(EULER(2)*EULER(4)+EULER(1)*EULER(3)),&
		 2.d0*(EULER(3)*EULER(4)-EULER(1)*EULER(2)),&
		 1.d0-2.d0*(EULER(2)**2+EULER(3)**2)/),(/3,3/))
END SUBROUTINE CAMAT
!
!***************************
SUBROUTINE SUMROT(Q,P,R)
!***************************
! Purpose: Calculates the resulting Euler parameters resulting from 2 consequtive rorations, Q and P.
!          The corresponding transformation matrix is CAMAT(R) = CAMAT(P).CAMAT(Q)
! Input
DOUBLE PRECISION    :: Q(4),P(4)   ! Euler parameters
! Output
DOUBLE PRECISION    :: R(4)        ! Transformation matrix
!
R(1) = Q(1)*P(1)  - DOT_PRODUCT(Q(2:4),P(2:4))
R(2:4) = P(1)*Q(2:4) + Q(1)*P(2:4) + CROSSPROD(P(2:4),Q(2:4))
!
END SUBROUTINE SUMROT
!
!***************************
SUBROUTINE SUMROTREL(Q,PQ,R)
!***************************
! Purpose: Calculates the resulting Euler parameters resulting from 2 consequtive rorations, Q and P(relative to Q)
!          Corresponds to "SUMROT", but here P is given in in Q-frame of reference
!          The corresponding transformation matrix is CAMAT(R) = CAMAT(TRANSPOSE(CAMAT(Q)).P).CAMAT(Q)
! Input
DOUBLE PRECISION    :: Q(4),PQ(4)   ! Euler parameters
! Output
DOUBLE PRECISION    :: R(4)        ! Transformation matrix
! Locals
DOUBLE PRECISION    :: AQ(3,3), P(4)
!
P(1) = PQ(1)
CALL CAMAT(Q,AQ)
!P(2:4) = MATMUL(TRANSPOSE(AQ),PQ(2:4))
P(2:4) = MATMUL(AQ,PQ(2:4))
!
R(1) = Q(1)*P(1)  - DOT_PRODUCT(Q(2:4),P(2:4))
R(2:4) = P(1)*Q(2:4) + Q(1)*P(2:4) + CROSSPROD(P(2:4),Q(2:4))
!
END SUBROUTINE SUMROTREL
!
!***************************
!***************************
SUBROUTINE AMAT2EULER(A, EULER)
!***************************
! Purpose: Calculates Euler parameters ("EULER") corresponding to then Transformation Matrix ("A")
! Euler parameters defined r0,rvec (r0,r1,r2,r3)
! input
DOUBLE PRECISION :: A(3,3)
! output
DOUBLE PRECISION :: EULER(4)
real*8 :: trace,S
! locals
!
! method from http://www.euclideanspace.com/maths/geometry/rotations/conversions/matrixToQuaternion/index.htm
!euler(1) = dsqrt( max( 0.0d0, 1 + A(1,1) + A(2,2) + A(3,3) ) ) / 2.0d0 
!euler(2) = dsqrt( max( 0.0d0, 1 + A(1,1) - A(2,2) - A(3,3) ) ) / 2.0d0 
!euler(3) = dsqrt( max( 0.0d0, 1 - A(1,1) + A(2,2) - A(3,3) ) ) / 2.0d0 
!euler(4) = dsqrt( max( 0.0d0, 1 - A(1,1) - A(2,2) + A(3,3) ) ) / 2.0d0 
!euler(2) = dsign( euler(2), A(3,2) - A(2,3) ) 
!euler(3) = dsign( euler(3), A(1,3) - A(3,1) ) 
!euler(4) = dsign( euler(4), A(2,1) - A(1,2) ) 

! other method from http://www.euclideanspace.com/maths/geometry/rotations/conversions/matrixToQuaternion/index.htm
trace=A(1,1)+A(2,2)+A(3,3)
if (trace>0.d0) then
  S=dsqrt(trace+1.d0)*2.d0 
  euler(1)=0.25d0*S
  euler(2)=(A(3,2)-A(2,3))/S
  euler(3)=(A(1,3)-A(3,1))/S
  euler(4)=(A(2,1)-A(1,2))/S
else if ((A(1,1).GT.A(2,2)).AND.(A(1,1).GT.A(3,3))) then
  S=dsqrt(1.d0+A(1,1)-A(2,2)-A(3,3))*2.d0
  euler(1)=(A(3,2)-A(2,3))/S
  euler(2)=0.25d0*S
  euler(3)=(A(1,2)+A(2,1))/S
  euler(4)=(A(1,3)+A(3,1))/S
else if (A(2,2).GT.A(3,3)) then
  S=dsqrt(1.d0-A(1,1)+A(2,2)-A(3,3))*2.d0
  euler(1)=(A(1,3)-A(3,1))/S
  euler(2)=(A(1,2)+A(2,1))/S
  euler(3)=0.25d0*S
  euler(4)=(A(2,3)+A(3,2))/S
else
  S=dsqrt(1.d0-A(1,1)-A(2,2)+A(3,3))*2.d0
  euler(1)=(A(2,1)-A(1,2))/S
  euler(2)=(A(1,3)+A(3,1))/S
  euler(3)=(A(2,3)+A(3,2))/S
  euler(4)=0.25d0*S
endif

END SUBROUTINE
!
!************************************
SUBROUTINE AMAT2EULER_SCALE(A, EULER)
!************************************
! Purpose: Calculates Euler parameters ("EULER") corresponding to then Transformation Matrix ("A")
! Same as "AMAT2EULER", but includes scaling of A matrix in order to handle small rotations
! input
DOUBLE PRECISION :: A(3,3)
! output
DOUBLE PRECISION :: EULER(4)
! locals
DOUBLE PRECISION :: ASCAL(3,3)
DOUBLE PRECISION :: lgt
INTEGER          :: i
real*8			 :: trace,S
! 
! Scale
DO i=1,3
  lgt = vec_length(A(1:3,i))
  ASCAL(1:3,i) = A(1:3,i)/lgt
ENDDO

!euler(1) = dsqrt( max( 0.0d0, 1 + A(1,1) + A(2,2) + A(3,3) ) ) / 2.0d0 
!euler(2) = dsqrt( max( 0.0d0, 1 + A(1,1) - A(2,2) - A(3,3) ) ) / 2.0d0 
!euler(3) = dsqrt( max( 0.0d0, 1 - A(1,1) + A(2,2) - A(3,3) ) ) / 2.0d0 
!euler(4) = dsqrt( max( 0.0d0, 1 - A(1,1) - A(2,2) + A(3,3) ) ) / 2.0d0 
!euler(2) = dsign( euler(2), A(3,2) - A(2,3) ) 
!euler(3) = dsign( euler(3), A(1,3) - A(3,1) ) 
!euler(4) = dsign( euler(4), A(2,1) - A(1,2) ) 

! other method from http://www.euclideanspace.com/maths/geometry/rotations/conversions/matrixToQuaternion/index.htm
trace=A(1,1)+A(2,2)+A(3,3)
if (trace>0.d0) then
  S=dsqrt(trace+1.d0)*2.d0 
  euler(1)=0.25d0*S
  euler(2)=(A(3,2)-A(2,3))/S
  euler(3)=(A(1,3)-A(3,1))/S
  euler(4)=(A(2,1)-A(1,2))/S
else if ((A(1,1).GT.A(2,2)).AND.(A(1,1).GT.A(3,3))) then
  S=dsqrt(1.d0+A(1,1)-A(2,2)-A(3,3))*2.d0
  euler(1)=(A(3,2)-A(2,3))/S
  euler(2)=0.25d0*S
  euler(3)=(A(1,2)+A(2,1))/S
  euler(4)=(A(1,3)+A(3,1))/S
else if (A(2,2).GT.A(3,3)) then
  S=dsqrt(1.d0-A(1,1)+A(2,2)-A(3,3))*2.d0
  euler(1)=(A(1,3)-A(3,1))/S
  euler(2)=(A(1,2)+A(2,1))/S
  euler(3)=0.25d0*S
  euler(4)=(A(2,3)+A(3,2))/S
else
  S=dsqrt(1.d0-A(1,1)-A(2,2)+A(3,3))*2.d0
  euler(1)=(A(2,1)-A(1,2))/S
  euler(2)=(A(1,3)+A(3,1))/S
  euler(3)=(A(2,3)+A(3,2))/S
  euler(4)=0.25d0*S
endif

END SUBROUTINE
!
!***************************
SUBROUTINE EV2EP(V,EULER)
!***************************
! Purpose: Calculates Euler parameters ("EULER") corresponding to then Euler angles ("V")
!          "V" is defined as rotation around z, y, x-axes, succesively
! input
DOUBLE PRECISION   :: V(3)
! output
DOUBLE PRECISION   :: EULER(4)
! locals
DOUBLE PRECISION   :: AMAT(3,3)
DOUBLE PRECISION   :: C(3),S(3)
!
C = dCOS(V)
S = dSIN(V)
!
AMAT = TRANSPOSE(RESHAPE((/&
       C(1)*C(2),-C(2)*S(1),S(2),&
       C(3)*S(1) + C(1)*S(2)*S(3), C(1)*C(3) - S(1)*S(2)*S(3) ,-C(2)*S(3),&
	  -C(1)*C(3)*S(2) + S(1)*S(3), C(3)*S(1)*S(2) + C(1)*S(3),C(2)*C(3)/),(/3,3/)))
!
CALL AMAT2EULER(AMAT,EULER)
!
!CHECK
!CALL CAMAT(EULER,A)

END SUBROUTINE
!
subroutine eaa2ep(v,euler)
  ! transformation from axis and angle to euler parameters (quaternions)
  real*8,dimension(4),intent(in)    :: v       ! x,y,z,angle
  real*8,dimension(4),intent(inout) :: euler  ! euler parameters (r0,r1,r2,r3)

  euler(2:4)=dsin(0.5d0*v(4))*v(1:3)
  euler(1)=dcos(0.5d0*v(4))
end subroutine eaa2ep
!----------------------------------------------------------------------------------------------------------
subroutine ep2eaa(euler,v)
  ! transformation from euler parameters (quaternions) to axis and angle 
  ! http://www.euclideanspace.com/maths/geometry/rotations/conversions/quaternionToAngle/index.htm
  real*8,dimension(4)	 :: euler	! euler parameters (r0,r1,r2,r3)
  real*8,dimension(4)    :: v       ! x,y,z,angle
  real*8				 :: s
  real*8,parameter		 :: eps=1.E-20
  euler=vec_norm(euler)				! if euler>1 acos and sqrt will produce errors, this cant happen if quaternion is normalised
  v(4)=2.0d0*dacos(euler(1))
  s=dsqrt(1-euler(1)**2)			! assuming quaternion normalised then w is less than 1, so term always positive.
  if (s<eps) then					! test to avoid divide by zero, s is always positive due to sqrt
    v(1)=euler(2)					! if s close to zero then direction of axis not important
	v(2)=euler(3)
	v(3)=euler(4)
  else
    v(1)=euler(2)/s					! normalise axis
	v(2)=euler(3)/s
	v(3)=euler(4)/s
  endif
end subroutine ep2eaa
!----------------------------------------------------------------------------------------------------------
subroutine get_two_plane_vectors(n,r1,r2)
! subroutine that calculates two perpendicular vectors r1 and r2 whoose crossproduct is the vector n
! In other words the two vectors r1 and r2 forms a plane which has the normal vector n
real*8,dimension(3),intent(inout) :: n
real*8,dimension(3),intent(inout) :: r1,r2

real*8,dimension(3) :: nc       ! copy of n
integer             :: index    
real*8              :: max
integer             :: i

real*8,save :: eps=1E-7

nc=vec_norm(n)

! find the index with the largest component of nc
max=0.0d0
do i=1,3
  if (abs(nc(i))>max) then
    max=abs(nc(i))
	index=i
  endif
enddo

select case (index)
case (1)    ! nc has the max composant at index 1
  if ((abs(nc(2))<eps).and.(abs(nc(3))<eps)) then
    r1=[0.0d0,1.0d0,0.0d0]
  else if (abs(nc(2))>abs(nc(3))) then
    r1(1)=0.0d0
    r1(3)=1.0d0/(1.0d0+(nc(3)/nc(2))**2)
    r1(2)=-nc(3)*r1(3)/nc(2)
  else
    r1(1)=0.0d0
	r1(2)=1.0d0/(1.0d0+(nc(2)/nc(3))**2)
	r1(3)=-nc(2)*r1(2)/nc(3)
  endif
case (2)    ! nc has the max composant at index 2
  if ((abs(nc(1))<eps).and.(abs(nc(3))<eps)) then
    r1=[1.0d0,0.0d0,0.0d0]
  else if (abs(nc(1))>abs(nc(3))) then
    r1(2)=0.0d0
    r1(3)=1.0d0/(1.0d0+(nc(3)/nc(1))**2)
    r1(1)=-nc(3)*r1(3)/nc(1)
  else
    r1(2)=0.0d0
	r1(1)=1.0d0/(1.0d0+(nc(1)/nc(3))**2)
	r1(3)=-nc(1)*r1(1)/nc(3)
  endif
case (3)    ! nc has the max composant at index 3
  if ((abs(nc(1))<eps).and.(abs(nc(2))<eps)) then
    r1=[1.0d0,0.0d0,0.0d0]
  else if (abs(nc(1))>abs(nc(2))) then
    r1(3)=0.0d0
    r1(2)=1.0d0/(1.0d0+(nc(2)/nc(1))**2)
    r1(1)=-nc(2)*r1(2)/nc(1)
  else
    r1(3)=0.0d0
    r1(1)=1.0d0/(1.0d0+(nc(1)/nc(2))**2)
    r1(2)=-nc(1)*r1(1)/nc(2)
  endif
end select

r2=crossprod(nc,r1)

! ANMH WAS HERE.............
n = n/dsqrt(dot_product(n,n))
r1 = r1/dsqrt(dot_product(r1,r1))
r2 = r2/dsqrt(dot_product(r2,r2))
! ..........TO HERE

end subroutine get_two_plane_vectors

!--------------------------------------------------------------------

Function vec_length(vec)
real*8, dimension(:), intent(in) :: vec
real*8                           :: vec_length

real*8 :: sum
integer :: i

sum=0.0
do i=1,size(vec)
  sum=sum+vec(i)**2
enddo
vec_length=dsqrt(sum)
end function

!----------------------------------------------

Function vec_norm(vec)
real*8, dimension(:), intent(in) :: vec
real*8, dimension(size(vec)) :: vec_norm

real*8 :: length

length=vec_length(vec)
if (length>1.0d-9) then
  vec_norm(:)=vec(:)*1.0d0/length
else
  vec_norm=0.0d0
endif

end function

!--------------------------------------------------------------------
!--------------------------------------------------------------------

Function rotate(a,n,azi)
! Function that rotates the vector a around another vector n
! with the angle azi
! a and n both intersect the same base point of a
real*8,dimension(1:3),intent(in) :: a,n
real*8,intent(in)                :: azi
real*8,dimension(1:3)            :: rotate

rotate=n*(dot_product(n,a))+ &
       dcos(azi)*(a-n*(dot_product(n,a)))+ &
	   dsin(azi)*(crossprod(n,a))
end function rotate

!--------------------------------------------------------------------
!--------------------------------------------------------------------

Function rotate_mat(a,n,azi)
! Function that rotates the column vectors in matrix a around another vector n
! with the angle azi
real*8,dimension(3,3),intent(in)	:: a
real*8,dimension(3),intent(in)		:: n
real*8,intent(in)					:: azi
real*8,dimension(3,3)				:: rotate_mat

integer									:: i

do i=1,3
  rotate_mat(:,i)=rotate(a(:,i),n,azi)
enddo
end function rotate_mat

!--------------------------------------------------------------------
!--------------------------------------------------------------------

Function intpol(x,x0,x1,f0,f1)   ! Linear interpolation function                                     
  real*8 intpol
  real*8,intent(in):: x,x0,x1,f0,f1
  real*8 :: eps=1.0e-7
  if (abs(x0-x1)<eps) then    ! changed by tjul 16.08.2006 to avoid division by zero
    intpol=f0
  else
    intpol=(x-x1)/(x0-x1)*f0+(x-x0)/(x1-x0)*f1
  endif
end function intpol

!--------------------------------------------------------------------
!--------------------------------------------------------------------

Function intpol_v(x,x0,x1,f0,f1)   ! Linear interpolation between one dim arrays
  real*8,intent(in)             :: x,x0,x1   
  real*8,dimension(:),intent(in):: f0,f1
  real*8,dimension(1:size(f0))  :: intpol_v

  integer i
  do i=1,size(f0)
    intpol_v(i)=intpol(x,x0,x1,f0(i),f1(i))
  enddo
end function intpol_v

!--------------------------------------------------------------------
!--------------------------------------------------------------------

SUBROUTINE EV2AMAT(V,A)
! Transformation matrix defined via Euler angles defined as:
!   1: V(1) around z, then
!   2: V(2) around y, and finally
!   3: V(3) around x
DOUBLE PRECISION :: V(3)
DOUBLE PRECISION :: A(3,3)
!
A = RESHAPE((/COS(V(1))*COS(V(2))                     , SIN(V(1))*COS(V(2))                     ,-SIN(V(2)),        &
     -SIN(V(1))*COS(V(3))+COS(V(1))*SIN(V(2))*SIN(V(3)), COS(V(1))*COS(V(3))+SIN(V(1))*SIN(V(2))*SIN(V(3)), COS(V(2))*SIN(V(3)), &
	  SIN(V(1))*SIN(V(3))+COS(V(1))*SIN(V(2))*COS(V(3)),-COS(V(1))*SIN(V(3))+SIN(V(1))*SIN(V(2))*COS(V(3)), COS(V(2))*COS(V(3))  /),(/3,3/))
!
END SUBROUTINE
!
!**********************
SUBROUTINE EV2AMAT_2(V,A)
!**********************
! Similar to EV2AMAT above, HOWEVER, the order on the parameter list is interchanged so that:
!   1: V(1) around x
!   2: V(2) around y
!   3: V(3) around z
! Still the rotation ORDER is the same, i.e. first z, then y, and finally x
DOUBLE PRECISION :: V(3), CV(3)
DOUBLE PRECISION :: A(3,3)
!
CV(1) = V(3)
CV(2) = V(2)
CV(3) = V(1)
CALL EV2AMAT(CV,A)
!
END SUBROUTINE
!
!**********************
SUBROUTINE AMAT2EV(A,V)
!**********************
! Calc. Euler angles from transformation matrix
! Euler angles are defined as:
!   1: V(1) around z, then
!   2: V(2) around y, and finally
!   3: V(3) around x
! 
DOUBLE PRECISION :: A(3,3)
DOUBLE PRECISION :: V(3)
!
V(1) = DATAN(A(2,1)/A(2,2))
V(2) = DASIN(-A(3,1))
V(3) = DATAN(A(3,2)/A(3,3))
!
END SUBROUTINE AMAT2EV
!
!**********************
SUBROUTINE AMAT2EV_2(A,V)
!**********************
! Calc. Euler angles from transformation matrix
! Euler angles are defined as:
!   1: V(1) around x, then
!   2: V(2) around y, and finally
!   3: V(3) around z
! Still the rotation ORDER is the same, i.e. first z, then y, and finally x
DOUBLE PRECISION :: A(3,3)
DOUBLE PRECISION :: V(3)
!
V(1) = DATAN(A(3,2)/A(3,3))
V(2) = DASIN(-A(3,1))
V(3) = DATAN(A(2,1)/A(2,2))
!
END SUBROUTINE AMAT2EV_2
END MODULE